<!-- begin headers -->
<!-- Load all dependencies -->
<script src="https://cdnjs.cloudflare.com/ajax/libs/pako/2.0.4/pako.min.js"></script>

<!-- JS scripts -->
<script src="{{static_path}}/chemiscope.min.js"></script>
<script src="{{static_path}}/chemiscope-sphinx.js"></script>

<!-- CSS styles -->
<link rel="stylesheet" href="{{static_path}}/chemiscope-sphinx.css" type="text/css" />
<!-- end headers -->

<!-- Warning Display -->
<div id="{{div_id}}-warning-display" class="chemiscope-sphinx-warning">
    <p></p>
    <button type="button" aria-label="Close" onclick="hideElement('{{div_id}}-warning-display')">
        &times;
    </button>
</div>
<div style="padding: 0.5rem 1.25rem" />

<!-- Loading Spinner -->
<div id="{{div_id}}-loading" class="chemiscope-sphinx-spinner">
    <img src="{{static_path}}/chemiscope-loading.svg" alt="Loading icon" />
</div>

<!-- Chemiscope Visualization -->
<div id="{{div_id}}">
    <!-- This will be replaced by the chemiscope HTML -->
</div>

<!-- Load Visualization Script -->
<script>
    loadChemiscopeSphinx('{{div_id}}', '{{filepath}}', '{{mode}}', '{{timeout}}');
</script>
