.chemiscope-sphinx-warning {
    display: none;
    flex-wrap: wrap;
    border: 1px solid #f0ad4e;
    background-color: #fcf8e3;
    color: #8a6d3b;
    padding: 15px;
    border-radius: 4px;
    margin-bottom: 20px;

    button {
        position: relative;
        float: right;
        font-size: 3em;
        line-height: 1;
        color: inherit;
        text-shadow: none;
        background-color: transparent;
        border: 0;
    }

    p {
        flex: 1;
    }
}

.chemiscope-sphinx-spinner img {
    width: 2em;
    height: 2em;
    animation: spin 2s linear infinite;
    position: absolute;
    z-index: 100000;
    left: 50%;
}

@keyframes spin {
    from {
        transform: rotate(0deg);
    }

    to {
        transform: rotate(360deg);
    }
}

.chemiscope-sphinx {
    .visualizer-container {
        display: flex;
        flex-wrap: wrap;
        background: white;
        padding-bottom: 16px;
        height: 550px;
    }

    .visualizer-column {
        flex: 1;
        position: relative;
        display: flex;
        flex-direction: column;
        padding-left: 10px;
        height: 100%;
    }

    .visualizer-column-right {
        flex: 1;
        position: relative;
        padding-right: 10px;
        height: 100%;
        display: flex;
        flex-direction: column;
    }

    /* ensure map container fills available space and allows toggle to be visible */
    .visualizer-column-right > div:last-child {
        flex: 1;
        min-height: 0;
        display: flex;
        flex-direction: column;
    }

    .visualizer-column-right > div:last-child > div:first-child {
        flex: 1;
        min-height: 0;
        height: auto !important;
    }

    .visualizer-item {
        width: 100%;
        height: 100%;
        flex: 1;
        min-height: 0;
    }

    .visualizer-info {
        flex-shrink: 0;
    }

    .visualizer-structure-mode,
    .visualizer-map-mode {
        margin: auto;
        max-width: 500px;
        height: 550px;
    }

    .visualizer-map-mode {
        display: block;
    }
}
